import functools
import six

from tensorflow import cast
from tensorflow import Variable
from tensorflow import math
from tensorflow import dtypes
from tensorflow import transpose
from tensorflow.random import uniform
from tensorflow import zeros_like
from tensorflow import concat

from tensorflow.signal import fft2d
from tensorflow.signal import rfft2d
from tensorflow.signal import ifft2d
from tensorflow.signal import irfft2d
from tensorflow.signal import fftshift
from tensorflow.signal import ifftshift

from tensorflow.keras.backend import permute_dimensions

from tensorflow.python.eager import context
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import activations
from tensorflow.python.keras import backend
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.engine.input_spec import InputSpec
# imports for backwards namespace compatibility
# pylint: disable=unused-import
# pylint: enable=unused-import
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import nn
from tensorflow.python.ops import nn_ops
from tensorflow.python.util.tf_export import keras_export
# pylint: disable=g-classes-have-attributes

"""
This class (ConvComplex2D) implements the complex convolution between two complex vectors (W = A + iB) and input vector (X = x + iy). The weights initialized in this class are complex and can be convolved with complex inputs such as fourier representation of image inputs...

Adapted from implementation of complex convolutions in Deep Complex Networks by Trabelsi et. al. (2018)

#ToDo: will use complex batch normalization
"""

class ConvComplex2D(Layer):
  """2-D complex convolution layer
  """

  def __init__(self,
               rank,
               filters,
               kernel_size,
               strides=1,
               padding='valid',
               data_format=None,
               dilation_rate=1,
               groups=1,
               activation=None,
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               trainable=True,
               name=None,
               conv_op=None,
               **kwargs):
    super(ConvComplex2D, self).__init__(
        trainable=trainable,
        name=name,
        activity_regularizer=regularizers.get(activity_regularizer),
        **kwargs)
    self.rank = rank

    if isinstance(filters, float):
        filters = int(filters)
    self.filters = filters
    self.groups = groups or 1
    self.kernel_size = conv_utils.normalize_tuple(
        kernel_size, rank, 'kernel_size')
    self.strides = conv_utils.normalize_tuple(strides, rank, 'strides')
    self.padding = conv_utils.normalize_padding(padding)
    self.data_format = conv_utils.normalize_data_format(data_format)
    self.dilation_rate = conv_utils.normalize_tuple(
        dilation_rate, rank, 'dilation_rate')

    self.activation = activations.get(activation)
    self.use_bias = use_bias

    self.kernel_initializer = initializers.get(kernel_initializer)
    self.bias_initializer = initializers.get(bias_initializer)
    self.kernel_regularizer = regularizers.get(kernel_regularizer)
    self.bias_regularizer = regularizers.get(bias_regularizer)
    self.kernel_constraint = constraints.get(kernel_constraint)
    self.bias_constraint = constraints.get(bias_constraint)
    self.input_spec = InputSpec(min_ndim=self.rank + 2)

    self._validate_init()
    self._is_causal = self.padding == 'causal'
    self._channels_first = self.data_format == 'channels_first'
    self._tf_data_format = conv_utils.convert_data_format(
        self.data_format, self.rank + 2)

  
  #kernel for spectral CNN
  def kernel_initialize(self, fan_in, fan_out):
    limit = math.sqrt(6/fan_in)
    return uniform(minval = -limit, maxval = limit, shape = [fan_in, fan_out, self.kernel_size[0], self.kernel_size[1]])
        
  def _validate_init(self):
    if self.filters is not None and self.filters % self.groups != 0:
        raise ValueError(
            'The number of filters must be evenly divisible by the number of '
            'groups. Received: groups={}, filters={}'.format(
                self.groups, self.filters))

    if not all(self.kernel_size):
        raise ValueError('The argument `kernel_size` cannot contain 0(s). '
                        'Received: %s' % (self.kernel_size,))

    if (self.padding == 'causal' and not isinstance(self, (Conv1D, SeparableConv1D))):
        raise ValueError('Causal padding is only supported for `Conv1D`'
                        'and `SeparableConv1D`.')

  def build(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape)
    self.input_channel = self._get_input_channel(input_shape)
    if self.input_channel % self.groups != 0:
            raise ValueError(
                'The number of input channels must be evenly divisible by the number '
                'of groups. Received groups={}, but the input has {} channels '
                '(full input shape is {}).'.format(self.groups, self.input_channel,
                                                    input_shape))
    
    
    self.kernel_shape = self.kernel_size  + (self.input_channel // 2 , self.filters*2)

    #fourier kernel (spectral)
    
    self.kernel = self.add_weight(
            name="kernel",
            shape=self.kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype)
    
    if self.use_bias:
        bias_shape = (self.filters*2,)
        self.bias = self.add_weight(
                name="bias",
                shape=bias_shape,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype)
    
        #self.bias = dtypes.complex(bias_fft, zeros_like(bias_fft), name='bias')
    else:
        self.bias = None
        
    #############################
    
    channel_axis = self._get_channel_axis()
    self.input_spec = InputSpec(min_ndim=self.rank + 2,
                                    axes={channel_axis: self.input_channel})

    # Convert Keras formats to TF native formats.
    if self.padding == 'causal':
        tf_padding = 'VALID'  # Causal padding handled in `call`.
    elif isinstance(self.padding, six.string_types):
        tf_padding = self.padding.upper()
    else:
        tf_padding = self.padding
    tf_dilations = list(self.dilation_rate)
    tf_strides = list(self.strides)

    
    tf_op_name = self.__class__.__name__
    '''
    if tf_op_name == 'Conv1D':
        tf_op_name = 'conv1d'  # Backwards compat.
    '''

    self._convolution_op = functools.partial(
        nn_ops.convolution_v2,
        strides=tf_strides,
        padding=tf_padding,
        dilations=tf_dilations,
        data_format=self._tf_data_format,
        name=tf_op_name)
    self.built = True

  def call(self, inputs):
    if self._is_causal:  # Apply causal padding to inputs for Conv1D.
            inputs = array_ops.pad(inputs, self._compute_causal_padding(inputs))
    
    #Complex convolution
    kernel_real = self.kernel[:,:,:,:self.filters]
    
    kernel_imag = self.kernel[:,:,:,self.filters:]
    
    cat_kernels_4_real = concat([kernel_real, -kernel_imag], axis=-2)
    cat_kernels_4_imag = concat([kernel_imag, kernel_real], axis=-2)
    cat_kernels_4_complex = concat(
            [cat_kernels_4_real, cat_kernels_4_imag], axis=-1
        )
    
    outputs = self._convolution_op(inputs, cat_kernels_4_complex)
    #print(outputs.shape)
    ###############
    
    if self.use_bias:
        output_rank = outputs.shape.rank
        if self.rank == 1 and self._channels_first:
        # nn.bias_add does not accept a 1D input tensor.
            bias = array_ops.reshape(self.bias, (1, self.filters, 1))
            outputs += bias
        else:
            # Handle multiple batch dimensions.
            if output_rank is not None and output_rank > 2 + self.rank:

                def _apply_fn(o):
                    return nn.bias_add(o, self.bias, data_format=self._tf_data_format)

                outputs = nn_ops.squeeze_batch_dims(
                    outputs, _apply_fn, inner_rank=self.rank + 1)
            else:
                outputs = nn.bias_add(
                    outputs, self.bias, data_format=self._tf_data_format)

    ################
    
    if self.activation is not None:
        return self.activation(outputs)

    return outputs


  def _spatial_output_shape(self, spatial_input_shape):
    return [
        conv_utils.conv_output_length(
            length,
            self.kernel_size[i],
            padding=self.padding,
            stride=self.strides[i],
            dilation=self.dilation_rate[i])
        for i, length in enumerate(spatial_input_shape)
    ]

  def compute_output_shape(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape).as_list()
    batch_rank = len(input_shape) - self.rank - 1
    if self.data_format == 'channels_last':
        return tensor_shape.TensorShape(
            input_shape[:batch_rank]
            + self._spatial_output_shape(input_shape[batch_rank:-1])
            + [self.filters])
    else:
        return tensor_shape.TensorShape(
            input_shape[:batch_rank] + [self.filters] +
            self._spatial_output_shape(input_shape[batch_rank + 1:]))

  def _recreate_conv_op(self, inputs):  # pylint: disable=unused-argument
    return False

  def get_config(self):
    config = {
        'filters':
             self.filters,
        'kernel_size':
            self.kernel_size,
        'strides':
            self.strides,
        'padding':
            self.padding,
        'data_format':
             self.data_format,
        'dilation_rate':
             self.dilation_rate,
        'groups':
            self.groups,
        'activation':
            activations.serialize(self.activation),
        'use_bias':
            self.use_bias,
        'kernel_initializer':
            initializers.serialize(self.kernel_initializer),
        'bias_initializer':
            initializers.serialize(self.bias_initializer),
        'kernel_regularizer':
            regularizers.serialize(self.kernel_regularizer),
        'bias_regularizer':
            regularizers.serialize(self.bias_regularizer),
        'activity_regularizer':
            regularizers.serialize(self.activity_regularizer),
        'kernel_constraint':
            constraints.serialize(self.kernel_constraint),
        'bias_constraint':
            constraints.serialize(self.bias_constraint)
    }
    base_config = super(ConvComplex2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def _compute_causal_padding(self, inputs):
    """Calculates padding for 'causal' option for 1-d conv layers."""
    left_pad = self.dilation_rate[0] * (self.kernel_size[0] - 1)
    if getattr(inputs.shape, 'ndims', None) is None:
         batch_rank = 1
    else:
         batch_rank = len(inputs.shape) - 2
    if self.data_format == 'channels_last':
        causal_padding = [[0, 0]] * batch_rank + [[left_pad, 0], [0, 0]]
    else:
        causal_padding = [[0, 0]] * batch_rank + [[0, 0], [left_pad, 0]]
    return causal_padding

  def _get_channel_axis(self):
    if self.data_format == 'channels_first':
        return -1 - self.rank
    else:
        return -1

  def _get_input_channel(self, input_shape):
    channel_axis = self._get_channel_axis()
    if input_shape.dims[channel_axis].value is None:
        raise ValueError('The channel dimension of the inputs '
                      'should be defined. Found `None`.')
    return int(input_shape[channel_axis])

  def _get_padding_op(self):
    if self.padding == 'causal':
        op_padding = 'valid'
    else:
        op_padding = self.padding
    if not isinstance(op_padding, (list, tuple)):
        op_padding = op_padding.upper()
    return op_padding




"""
This class (ConvSpectral2D) implements the spectral parametrization for learning weights in fourier space. In order to work around the complexity of complex convolution, the class levarage (Rippel et. al.) conjugate symmetry (irfft2d) for real valued signals...

#ToDo: 
"""

class ConvSpectral2D(Layer):
  """2-D spectral convolution layer
  """

  def __init__(self,
               rank,
               filters,
               kernel_size,
               strides=1,
               padding='valid',
               data_format=None,
               dilation_rate=1,
               groups=1,
               activation=None,
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               trainable=True,
               name=None,
               conv_op=None,
               **kwargs):
    super(ConvSpectral2D, self).__init__(
        trainable=trainable,
        name=name,
        activity_regularizer=regularizers.get(activity_regularizer),
        **kwargs)
    self.rank = rank

    if isinstance(filters, float):
        filters = int(filters)
    self.filters = filters
    self.groups = groups or 1
    self.kernel_size = conv_utils.normalize_tuple(
        kernel_size, rank, 'kernel_size')
    self.strides = conv_utils.normalize_tuple(strides, rank, 'strides')
    self.padding = conv_utils.normalize_padding(padding)
    self.data_format = conv_utils.normalize_data_format(data_format)
    self.dilation_rate = conv_utils.normalize_tuple(
        dilation_rate, rank, 'dilation_rate')

    self.activation = activations.get(activation)
    self.use_bias = use_bias

    self.kernel_initializer = initializers.get(kernel_initializer)
    self.bias_initializer = initializers.get(bias_initializer)
    self.kernel_regularizer = regularizers.get(kernel_regularizer)
    self.bias_regularizer = regularizers.get(bias_regularizer)
    self.kernel_constraint = constraints.get(kernel_constraint)
    self.bias_constraint = constraints.get(bias_constraint)
    self.input_spec = InputSpec(min_ndim=self.rank + 2)

    self._validate_init()
    self._is_causal = self.padding == 'causal'
    self._channels_first = self.data_format == 'channels_first'
    self._tf_data_format = conv_utils.convert_data_format(
        self.data_format, self.rank + 2)

  
  #kernel for spectral CNN
  def kernel_initialize(self, fan_in, fan_out):
    limit = math.sqrt(6/fan_in)
    return uniform(minval = -limit, maxval = limit, shape = [fan_in, fan_out, self.kernel_size[0], self.kernel_size[1]])
        
  def _validate_init(self):
    if self.filters is not None and self.filters % self.groups != 0:
        raise ValueError(
            'The number of filters must be evenly divisible by the number of '
            'groups. Received: groups={}, filters={}'.format(
                self.groups, self.filters))

    if not all(self.kernel_size):
        raise ValueError('The argument `kernel_size` cannot contain 0(s). '
                        'Received: %s' % (self.kernel_size,))

    if (self.padding == 'causal' and not isinstance(self, (Conv1D, SeparableConv1D))):
        raise ValueError('Causal padding is only supported for `Conv1D`'
                        'and `SeparableConv1D`.')

  def build(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape)
    self.input_channel = self._get_input_channel(input_shape)
    if self.input_channel % self.groups != 0:
            raise ValueError(
                'The number of input channels must be evenly divisible by the number '
                'of groups. Received groups={}, but the input has {} channels '
                '(full input shape is {}).'.format(self.groups, self.input_channel,
                                                    input_shape))
    
    
    self.kernel_shape = self.kernel_size  + (self.input_channel , self.filters*2)

    #fourier kernel (spectral)
    
    #print(self.kernel_shape)
    
    self.kernel_spatial = self.add_weight(
            name="kernel",
            shape=self.kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype)
    
    if self.use_bias:
        bias_shape = (self.filters,)
        self.bias = self.add_weight(
                name="bias",
                shape=bias_shape,
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype)
    
        #self.bias = dtypes.complex(bias_fft, zeros_like(bias_fft), name='bias')
    else:
        self.bias = None
        
    #############################
    
    channel_axis = self._get_channel_axis()
    self.input_spec = InputSpec(min_ndim=self.rank + 2,
                                    axes={channel_axis: self.input_channel})

    # Convert Keras formats to TF native formats.
    if self.padding == 'causal':
        tf_padding = 'VALID'  # Causal padding handled in `call`.
    elif isinstance(self.padding, six.string_types):
        tf_padding = self.padding.upper()
    else:
        tf_padding = self.padding
    tf_dilations = list(self.dilation_rate)
    tf_strides = list(self.strides)

    tf_op_name = self.__class__.__name__
    if tf_op_name == 'Conv1D':
        tf_op_name = 'conv1d'  # Backwards compat.

    self._convolution_op = functools.partial(
        nn_ops.convolution_v2,
        strides=tf_strides,
        padding=tf_padding,
        dilations=tf_dilations,
        data_format=self._tf_data_format,
        name=tf_op_name)
    self.built = True

  def call(self, inputs):
    if self._is_causal:  # Apply causal padding to inputs for Conv1D.
            inputs = array_ops.pad(inputs, self._compute_causal_padding(inputs))
    
    #Spectral CNN 
    
    kernel_spatial_trans = cast(transpose(self.kernel_spatial, [2, 3, 0, 1]), dtype=dtypes.float32)
    
    kernel = math.real(ifft2d(dtypes.complex(kernel_spatial_trans[:,:self.filters,:,:], kernel_spatial_trans[:,self.filters:,:,:])))
    
    kernel = cast(transpose(kernel, [2, 3, 0, 1]), dtype=self.dtype)
    
    outputs = self._convolution_op(inputs, kernel)
    #print(outputs.shape)
    ###############
    
    if self.use_bias:
        output_rank = outputs.shape.rank
        if self.rank == 1 and self._channels_first:
        # nn.bias_add does not accept a 1D input tensor.
            bias = array_ops.reshape(self.bias, (1, self.filters, 1))
            outputs += bias
        else:
            # Handle multiple batch dimensions.
            if output_rank is not None and output_rank > 2 + self.rank:

                def _apply_fn(o):
                    return nn.bias_add(o, self.bias, data_format=self._tf_data_format)

                outputs = nn_ops.squeeze_batch_dims(
                    outputs, _apply_fn, inner_rank=self.rank + 1)
            else:
                outputs = nn.bias_add(
                    outputs, self.bias, data_format=self._tf_data_format)

    ################
    
    if self.activation is not None:
        return self.activation(outputs)

    return outputs


  def _spatial_output_shape(self, spatial_input_shape):
    return [
        conv_utils.conv_output_length(
            length,
            self.kernel_size[i],
            padding=self.padding,
            stride=self.strides[i],
            dilation=self.dilation_rate[i])
        for i, length in enumerate(spatial_input_shape)
    ]

  def compute_output_shape(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape).as_list()
    batch_rank = len(input_shape) - self.rank - 1
    if self.data_format == 'channels_last':
        return tensor_shape.TensorShape(
            input_shape[:batch_rank]
            + self._spatial_output_shape(input_shape[batch_rank:-1])
            + [self.filters])
    else:
        return tensor_shape.TensorShape(
            input_shape[:batch_rank] + [self.filters] +
            self._spatial_output_shape(input_shape[batch_rank + 1:]))

  def _recreate_conv_op(self, inputs):  # pylint: disable=unused-argument
    return False

  def get_config(self):
    config = {
        'filters':
             self.filters,
        'kernel_size':
            self.kernel_size,
        'strides':
            self.strides,
        'padding':
            self.padding,
        'data_format':
             self.data_format,
        'dilation_rate':
             self.dilation_rate,
        'groups':
            self.groups,
        'activation':
            activations.serialize(self.activation),
        'use_bias':
            self.use_bias,
        'kernel_initializer':
            initializers.serialize(self.kernel_initializer),
        'bias_initializer':
            initializers.serialize(self.bias_initializer),
        'kernel_regularizer':
            regularizers.serialize(self.kernel_regularizer),
        'bias_regularizer':
            regularizers.serialize(self.bias_regularizer),
        'activity_regularizer':
            regularizers.serialize(self.activity_regularizer),
        'kernel_constraint':
            constraints.serialize(self.kernel_constraint),
        'bias_constraint':
            constraints.serialize(self.bias_constraint)
    }
    base_config = super(ConvSpectral2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def _compute_causal_padding(self, inputs):
    """Calculates padding for 'causal' option for 1-d conv layers."""
    left_pad = self.dilation_rate[0] * (self.kernel_size[0] - 1)
    if getattr(inputs.shape, 'ndims', None) is None:
         batch_rank = 1
    else:
         batch_rank = len(inputs.shape) - 2
    if self.data_format == 'channels_last':
        causal_padding = [[0, 0]] * batch_rank + [[left_pad, 0], [0, 0]]
    else:
        causal_padding = [[0, 0]] * batch_rank + [[0, 0], [left_pad, 0]]
    return causal_padding

  def _get_channel_axis(self):
    if self.data_format == 'channels_first':
        return -1 - self.rank
    else:
        return -1

  def _get_input_channel(self, input_shape):
    channel_axis = self._get_channel_axis()
    if input_shape.dims[channel_axis].value is None:
        raise ValueError('The channel dimension of the inputs '
                      'should be defined. Found `None`.')
    return int(input_shape[channel_axis])

  def _get_padding_op(self):
    if self.padding == 'causal':
        op_padding = 'valid'
    else:
        op_padding = self.padding
    if not isinstance(op_padding, (list, tuple)):
        op_padding = op_padding.upper()
    return op_padding


import sys
import numpy as np

import tensorflow as tf
import tensorflow_addons as tfa

#Custom libraries - User defined
from ..utils import conv_utils
from ..layers import poolSpectral
from ..utils.data_utils import parseDataset
from ..utils.conv_utils import convSpec2d_block, conv2d_block
from ..utils.conv_utils import cross_correlate_iff, cross_correlate_ff, cross_correlate

def complexUnetMultiHead(input_channel, out_channel, image_size, filter_size, n_depth, dp_rate = 0.1, act = 'relu'):
    
    input_size = (image_size, image_size, input_channel // 2)
    inputsA = tf.keras.Input(shape=input_size)   # input for CBED
    inputsB = tf.keras.Input(shape=input_size)   # input for probe
    
    output_size = (image_size, image_size, out_channel)
    true_out = tf.keras.Input(shape=output_size)   # input for probe

    skips = []

    print("Cross correlation layer ... \n")
    cc1 = tf.keras.layers.Lambda(cross_correlate_ff)([inputsA, inputsB])
    #cc1 = tf.keras.layers.BatchNormalization()(cc1)
    
    print("\n")
    print("Building the comlpex U-net ... \n")
    
    pool = convSpec2d_block(cc1, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act) 
    
    # conv u-net for 4DSTEM cbed pattern
    for encode in range(int(np.log2(image_size))-1):
        conv = convSpec2d_block(pool, np.minimum(filter_size * (2**encode), 256), n_depth=n_depth, dp_rate=dp_rate, activation=act)
        pool = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv)
        skips.append(conv)
        
    poolVg = pool
    poolQz= pool
    
    for decode in reversed(range(int(np.log2(image_size))-1)):
        convVg = convSpec2d_block(poolVg, np.minimum(filter_size * (2**(decode+1)), 256), n_depth=n_depth, dp_rate=dp_rate, activation=act)
        poolVg = tf.keras.layers.UpSampling2D(size=(2, 2))(convVg)
        poolVg = tf.keras.layers.Concatenate(axis=-1)([skips[decode], poolVg])
        
    for decode in reversed(range(int(np.log2(image_size))-1)):
        convQz = convSpec2d_block(poolQz, np.minimum(filter_size * (2**(decode+1)), 256), n_depth=n_depth, dp_rate=dp_rate, activation=act)
        poolQz = tf.keras.layers.UpSampling2D(size=(2, 2))(convQz)
        poolQz = tf.keras.layers.Concatenate(axis=-1)([skips[decode], poolQz])

    unetVg = convSpec2d_block(poolVg, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act) 
    unetQz = convSpec2d_block(poolQz, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act) 

    print("\n")
    print("Building the inverse fft layer ... \n")
    unet_real_vg = tf.keras.layers.Lambda(cross_correlate_iff, dtype='float32')(unetVg)
    unet_real_qz = tf.keras.layers.Lambda(cross_correlate_iff, dtype='float32')(unetQz)
    
    print("\n")
    print("Building the one more spatial convolution layers ... \n")
    conv_real_vg = conv2d_block(unet_real_vg, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act)
    conv_real_qz = conv2d_block(unet_real_qz, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act)

    print("\n")
    print("Building the final layer ... \n")
    pot_out = tf.keras.layers.Conv2D(filters = 1, kernel_size = (3, 3),padding='same', kernel_initializer = 'he_uniform')(conv_real_vg)
    qz_out = tf.keras.layers.Conv2D(filters = 1, kernel_size = (3, 3),padding='same', kernel_initializer = 'he_uniform')(conv_real_qz)
    
    pred_out = tf.keras.layers.Concatenate(axis=-1)([pot_out, qz_out])
    pred_out, losses = multiTaskLayer(scale_factor=0.001)([true_out, pred_out])
    pred_out = tf.keras.layers.Activation('linear', dtype='float32')(pred_out)
    
    model = tf.keras.Model(inputs=[inputsA, inputsB, true_out], outputs=(pred_out, losses))

    return model


import numpy as np

import functools
import six

from tensorflow import Variable
from tensorflow import math
from tensorflow import dtypes
from tensorflow import transpose
from tensorflow.random import uniform
from tensorflow import zeros_like
from tensorflow import ones_like
from tensorflow import concat

from tensorflow import int32
from tensorflow import float32
from tensorflow import cast
from tensorflow import concat
from tensorflow import constant
from tensorflow import less_equal
from tensorflow import transpose
from tensorflow import complex64
from tensorflow import expand_dims
from tensorflow import math
from tensorflow import dtypes
from tensorflow import divide
from tensorflow import multiply
from tensorflow.keras.initializers import random_uniform
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import backend
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.engine.input_spec import InputSpec
from tensorflow.python.util.tf_export import keras_export

from tensorflow.python.eager import context
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers

class multiTaskLayer(Layer):
    def __init__(self, sigma_size = 2,
                sigma_initializer = None,
                scale_factor = 0.1,
                sigma_penalty = 0.0001):
        super(multiTaskLayer, self).__init__()
        self.sigma_size = sigma_size
        self.scale_factor = scale_factor
        self.sigma_shape = [self.sigma_size,]
        self.sigma_penalty = sigma_penalty
        self._scale_factor = constant([scale_factor,0])
        if sigma_initializer:
            self.sigma_initializer = sigma_initializer
        else:
            self.sigma_initializer = random_uniform(minval=0.25, maxval=1.0)
            
                                                 
        self._sigma_sq = self.add_weight(
            name="sigma_sq",
            shape=self.sigma_shape,
            initializer=self.sigma_initializer,
            constraint=constraints.non_neg(),
            trainable=True,
            dtype=self.dtype)

    def call(self, inputs):
        assert(isinstance(inputs, list) or isinstance(inputs, tuple)), 'inputs to the layer should be either list or tuple of tensors'
        true_input = inputs[0]
        pred_input = inputs[1]
        
        assert(len(true_input.shape) == len(pred_input.shape)), 'true and prediction should have same dimension!!'
        true_input = cast(true_input, dtype=self._sigma_sq.dtype)
        pred_input = cast(pred_input, dtype=self._sigma_sq.dtype)
        
        #Multiply Vg to subsequent tasks (such as loss term for qz = square(true-pred)*true_vg)
        true_vg = expand_dims(true_input[:,:,:,0], axis=-1)
        padding = ones_like(true_vg, dtype=self._sigma_sq.dtype)
        true_vg = concat([padding, true_vg], axis=-1)
        
        self._sigma_sq_factor = multiply(cast(self._scale_factor, dtype=float32) 
                                         , divide(1.0 , multiply(2.0, (cast(self._sigma_sq, dtype=float32 )+ self.sigma_penalty))))
        
        self._log_sigma_sq = math.log(cast(self._sigma_sq, dtype=float32 )+ self.sigma_penalty)
        loss_tensor = multiply(self._sigma_sq_factor, math.square(pred_input-true_input)*true_vg) + self._log_sigma_sq
        losses = math.reduce_sum(loss_tensor, axis=-1)
        outputs = cast(pred_input, self.dtype)
        
        return outputs, losses
    
    def get_config(self):
        config = {
            'sigma_size':
            self.sigma_size,
            'sigma_initializer':
            initializers.serialize(self.sigma_initializer),
            'scale_factor':
            self.scale_factor,
            'sigma_penalty':
            self.sigma_penalty
        }
        base_config = super(multiTaskLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
        
        



import sys
import h5py
import numpy as np
from itertools import product
from time import sleep
import matplotlib.pyplot as plt

def plot_image(inputs, grid_size = [1,2], figsize = 20, scale=None, show_axis = False, show_title=False, save_fig=False, filename = 'test.png', title=None, **kwargs):
    assert(type(inputs) == list)
    assert(type(figsize) == int or tuple)
    assert(type(grid_size) == list)
    assert(len(inputs) == grid_size[0]*grid_size[1])
    
    if type(figsize) == int:
        figsize = (figsize, figsize)
        
    fig, axs = plt.subplots(grid_size[0], grid_size[1], figsize = figsize)
    
    counter = -1
    if grid_size[0] == 1:
        for i in range(grid_size[1]):
            counter += 1
            if scale and type(scale) == float:
                axs[i].imshow(inputs[counter]**scale, **kwargs)
            elif scale and type(scale) == str:
                assert(scale=='log'), "The scale type is not supported"
                axs[i].imshow(np.log(inputs[counter]), **kwargs)
            else:
                axs[i].imshow(inputs[counter], **kwargs)
                
            '''
            if show_title:
                assert(len[title] == grid_size[0]*grid_size[1]), 'titles for all subplots are not provided'
                axs[i].set_title(title[i])
            '''
            
            if not show_axis:
                axs[i].axes.xaxis.set_visible(False)
                axs[i].axes.yaxis.set_visible(False)
    else:
        for i in range(grid_size[0]):
            for j in range(grid_size[1]):
                counter += 1
                if scale and type(scale) == float:
                    axs[i][j].imshow(inputs[counter]**scale, **kwargs)
                elif scale and type(scale) == str:
                    assert(scale=='log'), "The scale type is not supported"
                    axs[i][j].imshow(np.log(inputs[counter]), **kwargs)
                else:
                    axs[i][j].imshow(inputs[counter], **kwargs)
                
                if not show_axis:
                    axs[i][j].axes.xaxis.set_visible(False)
                    axs[i][j].axes.yaxis.set_visible(False)
        plt.tight_layout()
        
    if save_fig:
        plt.savefig(filename, dpi=600)


def plot_line(x = None,y = None, xlabel = None, ylabel = None, figsize = 20, labelsize = 30, fontsize=30):
    assert(type(figsize) == int or tuple)
    if x is not None:
        assert(type(x) == np.ndarray or list)
    assert(type(y) == np.ndarray or list)
    
    if type(figsize) == int:
        figsize = (figsize, figsize)
        
    fig = plt.figure(figsize=figsize)

    if x is None:
        for i in range(len(y)):
            plt.plot(y[i])
    else:
        for i in range(len(y)):
            plt.plot(x, y[i])

    plt.tick_params(axis='both', which='major', labelsize=labelsize)
    if xlabel is not None:
        plt.xlabel(xlabel, fontsize = fontsize)
    if ylabel is not None:
        plt.ylabel(ylabel, fontsize = fontsize)
        
        
def plot_scatter(x =None, y=None, figsize = 20, scale=None, show_axis = False):
    plt.figure(figsize=(8, 8), dpi=80)

    plt.scatter(maxima_x,maxima_y, s = maxima_int, color='red', facecolors='none', linewidth=3, label = 'Predicted')
    plt.scatter(tr_maxima_x,tr_maxima_y, s = tr_maxima_int, color='blue', facecolors='none', linewidth=3, label = 'True')
    #plt.legend()
    plt.show()


import os
import sys
import h5py
import numpy as np
import tensorflow as tf

'''
This class is to load tensorflow records from the database
'''
# %%
# Read TFRecord file
class parseDataset(object):
    def __init__(self, filepath = '', image_size=256, out_channel=1):
        if os.path.isdir(filepath):
            files = []
            for filename in os.listdir(filepath):
                if filename.endswith(".h5") or filename.endswith(".hdf5"):
                    file = os.path.join(filepath, filename)
                    files.append(file)
                    self.ext = "hdf5"
                elif filename.endswith(".tfrecords"):
                    file = os.path.join(filepath, filename)
                    files.append(file)
                    self.ext = "tfrecords"
                else:
                    print("Currently only supports hdf5 or tfrecords as dataset \n")
                    exit()
            self.file_lists = files
            self.from_dir = True
        else:
            self.ext = filepath.split('.')[-1]
            assert(self.ext == 'hdf5' or self.ext == 'tfrecords'), "Currently only supports hdf5 or tfrecords as dataset" 
            self.filepath = filepath
            self.from_dir = False
            
        assert(isinstance(image_size, (int, list))), 'image_size must be integer (when height=width) or list (height, width)'
        if isinstance(image_size, int):
            self.height = image_size
            self.width = image_size
        else:
            self.height = image_size[0]
            self.width = image_size[1]
            
        self.out_channel = out_channel
        
    def read(self, batch_size = 128, shuffle = True, mode = 'default'):
        if self.ext == 'hdf5':
            ds = self.prepare_dataset_from_hdf5(batch_size , shuffle , mode)
        elif self.ext == 'tfrecords':
            ds = self.prepare_dataset_from_tfrecords(batch_size , shuffle , mode)
        return ds
            
    def prepare_dataset_from_tfrecords(self, batch_size , shuffle , mode):
        '''
        parse data from tfrecords
        '''
        #TODO: Scaling scheme and standardization
        if self.from_dir:
            tfr_dataset = tf.data.TFRecordDataset(self.file_lists) 
        else:
            tfr_dataset = tf.data.TFRecordDataset(self.filepath) 
    
        if mode == 'default':
            ds = tfr_dataset.map(self._parse_tfr_element)
        elif mode == 'norm':
            ds = tfr_dataset.map(self._parse_tfr_element_norm)
        elif mode == 'log':
            ds = tfr_dataset.map(self._parse_tfr_element_log)
        elif mode == 'multihead':
            ds = tfr_dataset.map(self._parse_tfr_element_multiHead)
        elif mode == 'multihead_norm':
            ds = tfr_dataset.map(self._parse_tfr_element_norm_multiHead)
        
    
        if shuffle:
            SHUFFLE_BUFFER_SIZE = 10000
            ds = ds.shuffle(SHUFFLE_BUFFER_SIZE)
        
        ds = ds.batch(batch_size, drop_remainder = True)
        
        return ds

    def prepare_dataset_from_hdf5(self, batch_size , shuffle , mode):
        '''
        parse data from hdf5
        '''
        data = h5py.File(self.filepath,'r')

        cbed = np.array(data.get('cbed').get('cbed_data'))
        probe = np.array(data.get('probe').get('probe_data'))
    
        if mode == 'default':
            pot = np.array(data.get('pot').get('pot_data'))
            ds = tf.data.Dataset.from_tensor_slices(((cbed,probe), pot))
        elif mode == 'norm':
            pot = np.array(data.get('pot').get('pot_data'))
            shape = pot.shape[0]
            pot = pot / np.amax(test_probe, axis = (1,2)).reshape(shape)
            ds = tf.data.Dataset.from_tensor_slices(((cbed,probe), pot))
        elif mode == 'multitask':
            pot = np.array(data.get('pot').get('pot_data'))
            shape = pot.shape[0]
            pot_max = np.amax(test_probe, axis = (1,2)).reshape(shape)
            pot = pot / pot_max
            ds = tf.data.Dataset.from_tensor_slices(((cbed,probe), (pot, pot_max)))
    
        if shuffle:
            SHUFFLE_BUFFER_SIZE = 10000
            ds = ds.shuffle(SHUFFLE_BUFFER_SIZE)
        
        ds = ds.batch(batch_size)
        
        return ds

    def _parse_tfr_element(self, element):
        parse_dic = {
            'cbed_feature': tf.io.FixedLenFeature([], tf.string), 
            'probe_feature': tf.io.FixedLenFeature([], tf.string),
            'pot_feature': tf.io.FixedLenFeature([], tf.string),
            }
        example_message = tf.io.parse_single_example(element, parse_dic)

        cbed_feature = example_message['cbed_feature'] 
        probe_feature = example_message['probe_feature']
        pot_feature = example_message['pot_feature']
        cbed = tf.io.parse_tensor(cbed_feature, out_type=tf.float32)
        cbed.set_shape([self.height, self.width, 1])
        probe = tf.io.parse_tensor(probe_feature, out_type=tf.float32)
        probe.set_shape([self.height, self.width, 1])
        pot = tf.io.parse_tensor(pot_feature, out_type=tf.float32)
        pot.set_shape([self.height, self.width, self.out_channel])
        
        return ((cbed,probe),self._replace_nan(pot))

    def _parse_tfr_element_norm(self, element):
        parse_dic = {
            'cbed_feature': tf.io.FixedLenFeature([], tf.string), 
            'probe_feature': tf.io.FixedLenFeature([], tf.string),
            'pot_feature': tf.io.FixedLenFeature([], tf.string),
            }
        example_message = tf.io.parse_single_example(element, parse_dic)

        cbed_feature = example_message['cbed_feature'] 
        probe_feature = example_message['probe_feature']
        pot_feature = example_message['pot_feature']
        cbed = tf.io.parse_tensor(cbed_feature, out_type=tf.float32)
        cbed.set_shape([self.height, self.width, 1])
        probe = tf.io.parse_tensor(probe_feature, out_type=tf.float32)
        probe.set_shape([self.height, self.width, 1])
        pot = tf.io.parse_tensor(pot_feature, out_type=tf.float32)
        pot.set_shape([self.height, self.width, self.out_channel])
        out_min = tf.reduce_min(pot, axis = [0,1])
        out_max = tf.reduce_max(pot, axis = [0,1])
        assert(out_min.shape == out_max.shape == self.out_channel)
        pot = (pot-out_min)/(out_max - out_min)
        
        return ((cbed,probe),self._replace_nan(pot))
    
    def _parse_tfr_element_log(self, element):
        parse_dic = {
            'cbed_feature': tf.io.FixedLenFeature([], tf.string), 
            'probe_feature': tf.io.FixedLenFeature([], tf.string),
            'pot_feature': tf.io.FixedLenFeature([], tf.string),
            }
        example_message = tf.io.parse_single_example(element, parse_dic)

        cbed_feature = example_message['cbed_feature'] 
        probe_feature = example_message['probe_feature']
        pot_feature = example_message['pot_feature']
        cbed = tf.io.parse_tensor(cbed_feature, out_type=tf.float32)
        cbed.set_shape([self.height, self.width, 1])
        probe = tf.io.parse_tensor(probe_feature, out_type=tf.float32)
        probe.set_shape([self.height, self.width, 1])
        pot = tf.io.parse_tensor(pot_feature, out_type=tf.float32)
        pot.set_shape([self.height, self.width, self.out_channel])
        pot = tf.math.log(pot)
        return ((cbed,probe),self._replace_nan(pot))
    
    def _parse_tfr_element_multiHead(self, element):
        '''
        Tensorflow multihead outputs (format: (probe,cbed),(vg,qz))
        Meant to be used for weighted loss implemented tensorflow
        '''
        parse_dic = {
            'cbed_feature': tf.io.FixedLenFeature([], tf.string), 
            'probe_feature': tf.io.FixedLenFeature([], tf.string),
            'pot_feature': tf.io.FixedLenFeature([], tf.string),
            }
        example_message = tf.io.parse_single_example(element, parse_dic)

        cbed_feature = example_message['cbed_feature'] 
        probe_feature = example_message['probe_feature']
        pot_feature = example_message['pot_feature']
        cbed = tf.io.parse_tensor(cbed_feature, out_type=tf.float32)
        cbed.set_shape([self.height, self.width, 1])
        probe = tf.io.parse_tensor(probe_feature, out_type=tf.float32)
        probe.set_shape([self.height, self.width, 1])
        pot = tf.io.parse_tensor(pot_feature, out_type=tf.float32)
        pot.set_shape([self.height, self.width, self.out_channel])
        
        pot_out = tf.expand_dims(pot[:,:,0], axis=-1)
        qz_out = tf.expand_dims(pot[:,:,1], axis=-1)
        
        return ((cbed,probe),(self._replace_nan(pot_out),self._replace_nan(qz_out)))
    
    def _parse_tfr_element_norm_multiHead(self, element):
        '''
        Tensorflow multihead normalized outputs (format: (probe,cbed),(vg,qz))
        Meant to be used for weighted loss implemented tensorflow
        '''
        parse_dic = {
            'cbed_feature': tf.io.FixedLenFeature([], tf.string), 
            'probe_feature': tf.io.FixedLenFeature([], tf.string),
            'pot_feature': tf.io.FixedLenFeature([], tf.string),
            }
        example_message = tf.io.parse_single_example(element, parse_dic)

        cbed_feature = example_message['cbed_feature'] 
        probe_feature = example_message['probe_feature']
        pot_feature = example_message['pot_feature']
        cbed = tf.io.parse_tensor(cbed_feature, out_type=tf.float32)
        cbed.set_shape([self.height, self.width, 1])
        probe = tf.io.parse_tensor(probe_feature, out_type=tf.float32)
        probe.set_shape([self.height, self.width, 1])
        pot = tf.io.parse_tensor(pot_feature, out_type=tf.float32)
        pot.set_shape([self.height, self.width, self.out_channel])
        out_min = tf.reduce_min(pot, axis = [0,1])
        out_max = tf.reduce_max(pot, axis = [0,1])
        assert(out_min.shape == out_max.shape == self.out_channel)
        pot = (pot-out_min)/(out_max - out_min)
        
        pot_out = tf.expand_dims(pot[:,:,0], axis=-1)
        qz_out = tf.expand_dims(pot[:,:,1], axis=-1)
        
        return ((cbed,probe),(self._replace_nan(pot_out),self._replace_nan(qz_out)))
    
    def _replace_nan(self,tensor):
        return tf.where(tf.math.is_nan(tensor), tf.zeros_like(tensor), tensor)


'''
=====================================================================================================================================
'''


"""
This is augmentation pipeline for 4DSTEM/STEM diffraction (CBED) images. The different available augmentation includes elliptic distrotion, plasmonic background noise and poisson shot noise. The elliptic distortion is recommeded to be applied on CBED, Probe, Potential Vg and Qz tilts while other noise are only applied on CBED input.
"""
import time
import numpy as np
import tensorflow as tf
import scipy.signal as sp
import numpy.matlib as nm
from itertools import product
from ..utils import aug_utils
from .interpolate import dense_image_warp
#from skimage.filters import gaussian
#from scipy.interpolate import griddata

from tensorflow.python.framework import tensor_shape

class image_augmentation(object):
    def __init__(self, 
                 backgrnd = True,
                 shot =True,
                 pattern_shift = True,
                 ellipticity = True,
                 counts_per_pixel = 1000,
                 qBackgroundLorentz = 0.1,
                 weightBackgrnd = 0.1,
                 scale = 1.0,
                 seed = None,
                 verbose = False,
                 log_file = './logs/augment_log.csv'):
        self.backgrnd = backgrnd
        self.shot = shot
        self.pattern_shift = pattern_shift
        self.ellipticity = ellipticity
        self.counts_per_pixel = counts_per_pixel
        self.qBackgroundLorentz = qBackgroundLorentz
        self.weightBackgrnd = weightBackgrnd
        self.scale = scale
        self.verbose = verbose
        self.log_file = log_file
        
        file_object = open(self.log_file, 'a')
        file_object.write('ellipticity,shot,background \n')
        file_object.close()

    def set_params(self, 
                 backgrnd = None, 
                 shot =None, 
                 pattern_shift = None,
                 ellipticity = None,
                 counts_per_pixel = None,
                 qBackgroundLorentz = None,
                 weightBackgrnd = None,
                 scale = None,
                 seed = None):
        
        if backgrnd is not None:
            self.backgrnd = backgrnd
        if shot is not None:
            self.shot = shot
        if pattern_shift is not None:
            self.pattern_shift = pattern_shift
        if ellipticity is not None:
            self.ellipticity = ellipticity
            
        if counts_per_pixel is not None:
            if self.shot:
                self.counts_per_pixel = counts_per_pixel
        if weightBackgrnd is not None and qBackgroundLorentz is not None:
            if self.backgrnd:
                self.weightBackgrnd = weightBackgrnd
                self.qBackgroundLorentz = qBackgroundLorentz
        if scale is not None:
            if self.ellipticity:
                self.scale = scale
        
    def get_params(self): 
        print('Printing augmentation summary... \n',end = "\r")
        print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n',end = "\r")
        print('Shots per pixel: {} \n'.format(self.counts_per_pixel),end = "\r")
        print('Background plasmon: {} \n'.format(self.qBackgroundLorentz),end = "\r")
        print('Ellipticity scaling: {} \n'.format(self.scale),end = "\r")
        print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< \n', end = "\r")
        
    
    def write_logs(self):
        file_object = open(self.log_file, 'a')
        file_object.write('{},{},{} \n'.format(self.ellipticity,self.shot,self.backgrnd))
        file_object.close()
        

    def _get_dim(self, x, idx):
        if x.shape.ndims is None:
            return tf.shape(x)[idx]
        return x.shape[idx] or tf.shape(x)[idx]    
    
    @tf.function
    def augment_img(self, inputs: aug_utils.TensorLike, probe = None):
        start_time = time.time()
        input_shape = inputs.shape
        if self.backgrnd:
            #TODO: Background plasmons
            input_noise = self.backgrnd_plasmon(inputs, probe)
        else:
            input_noise = inputs
            self.weightBackgrnd = 0
            self.qBackgroundLorentz = 0
        
        if self.shot:
            input_noise = self.shot_noise(input_noise)
        else:
            self.counts_per_pixel = 0
            
        if self.pattern_shift:
            #TODO: Pattern shift for cbed
            pass
        
        if self.ellipticity:
            #TODO: Implement elpticity for probe and cbed
            input_noise = self.elliptic_distort(input_noise)
        else:
            self.scale = 0
        
        t = time.time() - start_time
        t = t/60
        
        if self.verbose:
            self.get_params()
            self.write_logs()
            print('Augmentation Status: it took {} minutes to augment {} images... \n'.format(t, input_shape[0]), end = "\r")
        else:
            self.write_logs()
            
        return input_noise
    
    @tf.function
    def scale_image(self, inputs: aug_utils.TensorLike):
        '''
        Scale image between 0 and 1
        '''
        input_shape = tf.shape(inputs)
        mean = tf.squeeze(tf.math.reduce_sum(inputs, axis = (1,2)))
        inputs_scaled = tf.transpose(tf.transpose(inputs, [3,1,2,0])/mean, [3,1,2,0])
        
        return inputs_scaled
    
    @tf.function
    def shot_noise(self, inputs: aug_utils.TensorLike):
        #input_shape = inputs.shape
        image_scale = self.scale_image(inputs)
        image_noise = tf.random.poisson(shape = [], lam = tf.maximum(image_scale,0) * self.counts_per_pixel)/float(self.counts_per_pixel)
        
        return image_noise
    
    @tf.function
    def elliptic_distort(
        self, inputs: aug_utils.TensorLike):
        """
        Elliptic distortion
        """
        image = tf.convert_to_tensor(inputs)
        batch_size, height, width, channels = (
            self._get_dim(image, 0),
            self._get_dim(image, 1),
            self._get_dim(image, 2),
            self._get_dim(image, 3),
        )
        
        exx = 0.7 * self.scale
        eyy = -0.5 * self.scale
        exy = -0.9 * self.scale
        
        m = [[1+exx, exy/2], [exy/2, 1+eyy]]
        
        grid_x, grid_y = tf.meshgrid(tf.range(width), tf.range(height))
        grid_x = grid_x - tf.math.reduce_mean(grid_x)
        grid_y = grid_y - tf.math.reduce_mean(grid_y)
        
        grid_x = tf.cast(grid_x, tf.float32)
        grid_y = tf.cast(grid_y, tf.float32)
        
        #print(stacked_grid)

        flow_x =  (grid_x*m[0][0] + grid_y*m[1][0])
        flow_y =  (grid_x*m[0][1] + grid_y*m[1][1])
        
        flow_grid = tf.cast(tf.stack([flow_y, flow_x], axis=2), grid_x.dtype)
        batched_flow_grid = tf.expand_dims(flow_grid, axis=0)
        
        batched_flow_grid = tf.repeat(batched_flow_grid, batch_size, axis=0)
        #assert(self._get_dim(batched_flow_grid, 0) == batch_size); "The flow batch size and the image batch size is different!!"
        
        imageOut = dense_image_warp(image, batched_flow_grid)
       
        return imageOut
    
    @tf.function
    def backgrnd_plasmon(self, inputs: aug_utils.TensorLike, probe: aug_utils.TensorLike):
        """
        Apply background plasmon noise
        """
        image = tf.convert_to_tensor(inputs)
        batch_size, height, width, channels = (
            self._get_dim(image, 0),
            self._get_dim(image, 1),
            self._get_dim(image, 2),
            self._get_dim(image, 3),
        )
        
        qx,qy = self._get_qx_qy([height,width])
        CBEDbg = 1./ (qx**2 + qy**2 + self.qBackgroundLorentz**2)
        CBEDbg = tf.squeeze(CBEDbg)
        CBEDbg = CBEDbg / tf.math.reduce_sum(CBEDbg)

        CBEDbg = tf.expand_dims(CBEDbg, axis=0)
        CBEDbg = tf.repeat(CBEDbg, batch_size, axis=0)
        CBEDbg = tf.expand_dims(CBEDbg, axis=3)
        
        CBEDbg = tf.keras.backend.permute_dimensions(CBEDbg, (3,0,1,2))
        probe = tf.keras.backend.permute_dimensions(probe, (3,0,1,2))

        CBEDbg_ff = tf.cast(tf.signal.rfft2d(CBEDbg), tf.complex128)
        probe_ff = tf.cast(tf.signal.rfft2d(probe), tf.complex128)
        mul_ff = tf.multiply(CBEDbg_ff, probe_ff)
        
        CBEDbgConv = tf.signal.fftshift(tf.signal.irfft2d(mul_ff), axes = [2,3])
        CBEDbgConv = tf.keras.backend.permute_dimensions(CBEDbgConv, (1,2,3,0))
        
        CBEDout = tf.cast(inputs, tf.float32) * (1-self.weightBackgrnd) + tf.cast(CBEDbgConv, tf.float32) * self.weightBackgrnd
        
        return CBEDout
        
    def _get_qx_qy(self, input_shape, pixel_size_AA = 0.20):
        """
        get qx,qy from cbed
        """
        N = input_shape
        #pixel_size_AA = example.realspacePixelSizeX * 2
        qx = np.sort(np.fft.fftfreq(N[0], pixel_size_AA)).reshape((N[0], 1, 1))
        qy = np.sort(np.fft.fftfreq(N[1], pixel_size_AA)).reshape((1, N[1], 1))
        
        return qx, qy


#Copyright: Joydeep Munshi
#Date: 07/01/2021
import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize

def predictions(X, gpr):
    mu, sigma = gpr.predict(X, return_std=True)
    mu = np.exp(mu)
    return mu


def expected_improvement(X, X_sample, Y_sample, gpr, xi):
    
    ''' Computes the EI at points X based on existing samples X_sample and Y_sample using a Gaussian process surrogate model. 
    Args: X: Points at which EI shall be computed (m x d). X_sample: Sample locations (n x d). Y_sample: Sample values (n x 1). 
    gpr: A GaussianProcessRegressor fitted to samples. xi: Exploitation-exploration trade-off parameter. 
    Returns: Expected improvements at points X. '''
    
    mu, sigma = gpr.predict(X, return_std=True)
    mu_sample = gpr.predict(X_sample)

    #sigma = sigma.reshape(-1, 2)
    sigma = sigma.reshape(-1, 1)
    
    # Needed for noise-based model,
    # otherwise use np.max(Y_sample).
    # See also section 2.4 in [...]
    mu_sample_opt = np.min(mu_sample)
    #print(mu_sample_opt)

    with np.errstate(divide='warn'):
        imp = mu - mu_sample_opt - xi
        Z = imp / sigma
        ei = imp * norm.cdf(Z) + sigma * norm.pdf(Z)
        ei[sigma == 0.0] = 0.0

    return ei[0][0]


def propose_location(acquisition, X_sample, Y_sample, gpr, space, bounds, xi=0.01, n_restarts=500, method='L-BFGS-B'):
    
    ''' Proposes the next sampling point by optimizing the acquisition function. 
    Args: acquisition: Acquisition function. X_sample: Sample locations (n x d). 
    Y_sample: Sample values (n x 1). gpr: A GaussianProcessRegressor fitted to samples. 
    Returns: Location of the acquisition function maximum. '''
    
    dim = X_sample.shape[1]
    min_val = 10000
    min_x = None
    c = 0
    
    def min_obj(X):
        # Minimization objective is the negative acquisition function
        return acquisition(X.reshape(-1, dim), X_sample, Y_sample, gpr, xi)
    
    # Find the best optimum by starting from n_restart different random points.
    for i in range(n_restarts):
        c = c+1
        params = {}
        for key in space.keys():
            params[key] = space[key]()
        x0 = np.asarray(np.asarray(list(params.values())))
        
        res = minimize(min_obj, x0=x0, bounds=bounds, method=method)
        if res.fun < min_val:
            etol = min_val - res.fun
            min_val = res.fun
            min_x = res.x 
    X_pred = min_x
    muu,sig = gpr.predict(X_pred.reshape(-1, dim),return_std=True)
            
    return X_pred,muu


def best_location(acquisition,X_sample, gpr, space, bounds, n_restarts=50000, method='L-BFGS-B'):
    
    
    dim = X_sample.shape[1]
    min_val = 10000
    min_x = None
    c = 0
    
    def min_obj(X):
        # Minimization objective is the negative acquisition function
        return acquisition(X.reshape(-1, dim), gpr)
    
    # Find the best optimum by starting from n_restart different random points.
    for i in range(n_restarts):
        c = c+1
        params = {}
        for key in space.keys():
            params[key] = space[key]()
        x0 = np.asarray(np.asarray(list(params.values())))
        
        res = minimize(min_obj, x0=x0, bounds=bounds, method=method)     
        if res.fun < min_val:
            etol = min_val - res.fun
            min_val = res.fun
            min_x = res.x  
    X_pred = min_x
    muu,sig = gpr.predict(X_pred.reshape(-1, dim),return_std=True)
            
    return X_pred,muu


import os
import sys
import numpy as np
import tensorflow as tf
from tensorflow.python.keras.utils import losses_utils

'''
These are custom Loss functions
'''

class MultiTaskLoss(tf.keras.losses.Loss):
    def __init__(self, loss_num=2, 
                 scale_factor = 0.01,
                 lam = 0.00001,
                 sigmas_sq = None,
                 reduction=tf.keras.losses.Reduction.NONE,
                 reduction_type = None,
                 name='multiTaskMse',
                 **kwargs):
        super(MultiTaskLoss, self).__init__(reduction=reduction,name=name,
                                            **kwargs)
        self._loss_num = loss_num
        self._scale_factor = scale_factor
        self._lambda = lam
        self.reduction_type = reduction_type
        if sigmas_sq:
            self._sigmas_sq = tf.Variable(name = 'Sigma_sq_',     
                                          dtype=tf.float32, 
                                          trainable= True,
                                          validate_shape = False,
                                          initial_value = sigmas_sq)
        else:
            self._sigmas_sq = tf.Variable(name = 'Sigma_sq_',     
                                          dtype=tf.float32, 
                                          trainable= True,
                                          validate_shape = False,
                                          initial_value = tf.initializers.random_uniform(minval=0.25, maxval=1.0)(shape = [loss_num,]))

    
    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        assert(len(y_true.shape) == 4), 'channel dimension mismatch!! recieved shape as {}'.format(y_true.shape)
        assert(len(y_pred.shape) == 4), 'channel dimension mismatch!! recieved shape as {}'.format(y_pred.shape)
        assert(y_pred.shape[-1] == self._loss_num), 'Please check the number of tasks are not same as the number of channels in the prediction'
        assert(y_true.shape[-1] == y_pred.shape[-1]), 'Dimensions of true and predicted outputs do not match!! expected {} but recieved {}'.format(y_true.shape,y_pred.shape)
        self.batch_size = y_true.shape[0]
        
        factor = tf.multiply(self._scale_factor, tf.divide(1.0, tf.multiply(2.0, self._sigmas_sq[0])))
        l2 = tf.math.square(y_true[:,:,:,0] - y_pred[:,:,:,0])
        loss = tf.add(tf.multiply(factor, l2), tf.math.log(self._sigmas_sq[0]+self._lambda))
        for i in range(1, self._loss_num):
            factor = tf.divide(1.0, tf.multiply(2.0, self._sigmas_sq[i]))
            l2 = tf.multiply(tf.math.square(y_true[:,:,:,i] - y_pred[:,:,:,i]), y_true[:,:,:,0])
            loss = tf.add(loss, tf.add(tf.multiply(factor, l2), tf.math.log(self._sigmas_sq[i]+self._lambda)))
        
        if self.reduction_type == 'distributed':
            loss = tf.reduce_mean(loss, axis = [1,2])
            loss = tf.reduce_sum(loss)* (1. / self.batch_size)
            
        return loss
    
    def print_config(self):
        print('sigmas_sq: {}'.format(self._sigmas_sq))
        print('batch_size: {}'.format(self.batch_size))
    
    '''
    def get_config(self):
        """Returns the config dictionary for a `Loss` instance."""
        return {'reduction': self.reduction, 'name': self.name, 'loss_num': self._loss_num,
                'scale_factor':  self._scale_factor, 'lam':  self._lambda,
                'sigmas_sq': tf.io.serialize_tensor(self._sigmas_sq, name='sigma')}
    '''
    
 
class CustomL1Loss(tf.keras.losses.Loss):
    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        
        return tf.math.reduce_mean(tf.math.abs(y_true - y_pred), axis=-1)

class customSSIMLoss(tf.keras.losses.Loss):
    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        
        data_range = tf.math.reduce_max(y_true) - tf.math.reduce_min(y_true)
        ssim_loss = 1-tf.image.ssim(y_true, y_pred, max_val=data_range)
        
        return ssim_loss
    
class CustomSSIML1ELoss(tf.keras.losses.Loss):    
    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        
        data_range = tf.math.reduce_max(y_true) - tf.math.reduce_min(y_true)
        ssim_loss = 1-tf.image.ssim(y_true, y_pred, max_val=data_range)
        mse_loss = tf.math.reduce_mean(tf.math.abs(y_true - y_pred), axis=(1,2,3))
        
        loss = ssim_loss + mse_loss
        
        return loss
